import abc
import json
import os
import tempfile

from asset_utils.common import exceptions, Singleton
from .transport_resource import TransportResource

DEFAULT_TIMEOUT = 86400  # in seconds i.e. 24 hours (60 * 60 * 24)
DEFAULT_BATCH_SIZE = 400  # max 400 concurrent requests


class Transporter(Singleton):
    credentials: dict = None
    prefixes: list = None  # storage system prefixes

    @classmethod
    def shared(cls, credentials: dict = None,
               prefixes: list = None,
               validate: bool = True):
        # TODO: set validate to True while using in asset-server
        kwargs = {"credentials": credentials, "prefixes": prefixes}
        transport = super(Transporter, cls).shared(**kwargs)
        if not validate:
            return transport

        try:
            transport.validate(credentials=credentials)
        except exceptions.InvalidStorageCredentialsError:
            # re-initialize
            print(f"found cached credentials for transporter:{cls.__name__}, reinitializing")
            transport.post_init(credentials=credentials, prefixes=prefixes)
        return transport

    def post_init(self, **kwargs):
        self.credentials = kwargs.get("credentials")
        self.prefixes = kwargs.get("prefixes")

    def validate(self, credentials: dict):
        if not self.credentials:
            raise exceptions.AssetException("missing storage credentials")
        # verify that credentials are matching, otherwise it causes a bug when
        # server receives requests from two different projects that are in the same
        # storage back-end
        if json.dumps(credentials) != json.dumps(self.credentials):
            raise exceptions.InvalidStorageCredentialsError()

        # TODO: verify that the storage system is accessible
        # validation should verify credentials exist and credentials are valid
        # (try to check to if you can access the bucket), i.e. bucket.name

    @property
    def batch_size(self):
        """Returns the batch size for upload, download and copy transactions."""
        return int(os.getenv("ASSET_BATCH_SIZE")) if os.getenv("ASSET_BATCH_SIZE") else DEFAULT_BATCH_SIZE

    def transfer(self, resources: [TransportResource]):
        """Generic interface for transferring from one source to another.
        This is meant to be the first landing point for all upload, download or copy
        operations:
            - upload: if dst has prefix but src doesn't
            - download: if src has prefix but dst doesn't
            - copy: if src and dst both have the prefixes

        Inheriting transporter classes can implement logic to determine which sub operation
        to invoke
        """
        copies, downloads, uploads = [], [], []
        for resource in resources:
            src_is_native = self.is_native_url(resource.src)
            dst_is_native = self.is_native_url(resource.dst)
            if src_is_native and dst_is_native:
                copies.append(resource)
            elif src_is_native:
                # src is already inside the transporter's native system, so we need to download
                downloads.append(resource)
            elif dst_is_native:
                uploads.append(resource)
        if copies:
            self.copy(resources=copies)
        if downloads:
            self.download(resources=downloads)
        if uploads:
            self.upload(resources=uploads)

    def get_download_resource(self, src: str, dst: str, src_hash: tuple) -> TransportResource:
        raise NotImplementedError

    def get_upload_resource(self, src: str, dst: str, src_hash: tuple) -> TransportResource:
        raise NotImplementedError

    def get_copy_resource(self, src: str, dst: str, src_hash: tuple, **kwargs) -> TransportResource:
        raise NotImplementedError

    def is_native_url(self, url: str):
        for prefix in self.prefixes:
            if url.startswith(prefix):
                return True
        return False

    @abc.abstractmethod
    def upload(self, resources: [TransportResource]):
        """upload resources from src to dst"""
        raise NotImplementedError

    @abc.abstractmethod
    def download(self, resources: [TransportResource]):
        """download resources from src to dst"""
        raise NotImplementedError

    @abc.abstractmethod
    def copy(self, resources: [TransportResource]):
        """copy resources from src to dst"""
        raise NotImplementedError

    def timeout(self) -> int:
        """returns the timeout for upload and download transactions"""
        return DEFAULT_TIMEOUT

    # used in asset-server
    def write_to_bucket(self, data: [dict]) -> None:
        """Write data directly to the bucket.

        Parameters
        ----------
        data: list of dicts {file: <InMemoryFile>, url: str}
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            resources = []
            for item in data:
                file_path = os.path.join(temp_dir, os.path.basename(item["url"]))
                # write to temp file
                item["file"].write(path=file_path)
                resources.append(self.get_upload_resource(src=file_path,
                                                          dst=item["url"],
                                                          src_hash=tuple()))
            # upload to bucket
            self.upload(resources=resources)



import copy
import fnmatch
import os

from asset_utils.utils.file_utils import FileUtils
from asset_utils.utils.in_memory_zip import InMemoryZip


def filter_blobs(blobs: list,
                 name_key: str,
                 pattern: str = None,
                 ignore: [str] = None) -> list:
    """Filter the blobs based on the pattern and ignore.

    TODO: Use unix style pattern matching for filtering the blobs.
    """
    if ignore:
        # ignore_patterns = ignore.split(",")
        not_ignored = []
        for blob in blobs:
            if not _is_ignored(getattr(blob, name_key), ignore):
                not_ignored.append(blob)
    else:
        not_ignored = list(blobs)

    if pattern:
        result = [blob for blob in not_ignored if _is_matched(getattr(blob, name_key), pattern)]
    else:
        result = not_ignored
    return result


def _is_matched(path, pattern=None):
    if not pattern:
        return True
    return fnmatch.fnmatch(path, pattern)


def _is_ignored(path: str, ignores: [str] = None):
    if not ignores:
        return False
    for ignore in ignores:
        if fnmatch.fnmatch(path, ignore):
            return True
    return False


# used in asset-server
def parse_for_saving_to_bucket(data: list):
    parsed = copy.deepcopy(data)
    for item in parsed:
        item["data"] = serialize_for_file(filename=item["url"], filedata=item["data"])
    return parsed


# used in asset-server
def serialize_for_file(filename: str, filedata):
    _, ext = os.path.splitext(filename)
    if ext in [".yaml", ".yml"]:
        return FileUtils.yaml_serialize(filedata)
    elif ext in [".json"]:
        return FileUtils.json_serialize(filedata)
    elif ext in [".zip"]:
        memzip = InMemoryZip()
        serialized = []
        for filename in filedata:
            # convert recursively
            serialized.append(serialize_for_file(filename, filedata[filename]))
        memzip.add_files(serialized)
        return memzip
    else:
        return filedata



import re

from asset_utils.common import exceptions


class MountConfig:
    def __init__(self, mount: str, url: str):
        self.mount = mount
        self.url = url

    def __str__(self):
        return f"MountConfig(mount={self.mount}, url={self.url})"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.mount == other.mount and self.url == other.url

    def __hash__(self):
        return hash((self.mount, self.url))

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def sanitize_path(path: str) -> str:
        """Remove any extra slashes from the path while preserving protocol identifiers."""
        # Preserve the protocol identifier if present
        protocol_match = re.match(r'^(\w+:\/\/)', path)
        protocol = protocol_match.group(1) if protocol_match else ''

        # Remove the protocol for now
        path_without_protocol = re.sub(r'^\w+:\/\/', '', path)

        # Replace multiple slashes with a single slash
        sanitized_path = re.sub(r'\/+', '/', path_without_protocol)

        # Reattach the protocol
        return protocol + sanitized_path

    def url_to_posix(self, url: str) -> str:
        """Translates the given URL to the mounted path"""
        if not self.is_url(url):
            raise exceptions.AssetException(f"URL {url} does not start with {self.url}")
        return self.sanitize_path(url.replace(self.url, self.mount))

    def posix_to_url(self, posix: str) -> str:
        """Translates the given mounted path to the URL"""
        if not self.is_posix(posix):
            raise exceptions.AssetException(f"Posix path {posix} does not start with {self.mount}")
        return self.sanitize_path(posix.replace(self.mount, self.url))

    def is_posix(self, path: str) -> bool:
        """Checks if the given path is a mounted path"""
        return path.startswith(self.mount)

    def is_url(self, url: str) -> bool:
        """Checks if the given URL is a mounted URL"""
        return url.startswith(self.url)



import pluggy

from asset_pluggy.plugin import APP_NAME
from asset_pluggy.plugin import specs as plugin_specs
from asset_utils.common.singleton import Singleton


class PluginManager(Singleton):

    def post_init(self):
        self._manager = pluggy.PluginManager(APP_NAME)
        for plugin in plugin_specs.__all__:
            self._manager.add_hookspecs(plugin)
        self._manager.load_setuptools_entrypoints(APP_NAME)

    @property
    def hook(self):
        return self._manager.hook

    def register(self, plugin):
        self._manager.register(plugin)



from asset_pluggy.storage.storage_credentials import StorageCredentials


def test_credentials_property():
    storage_credentials = StorageCredentials.shared()
    content_credentials = {"user": "content_user", "password": "test123"}
    project_credentials = {"user": "project_user", "password": "admin123"}
    # set both content and project credentials
    storage_credentials.set_content_credentials(content_credentials)
    storage_credentials.set_credentials(project_credentials)
    # check both credentials
    assert storage_credentials.credentials == project_credentials
    storage_credentials.use_content_credentials = True
    assert storage_credentials.credentials == content_credentials



from __future__ import annotations

import abc
import os
from typing import Any

from asset_pluggy.storage.urls import StorageURL
from asset_utils.common import exceptions
from asset_utils.utils.log_utils import LoggingMixin

PATH_SEPARATOR = "/"


class StorageData(LoggingMixin, metaclass=abc.ABCMeta):
    url: str
    host: str
    name: str
    size: int = None
    content_type: str = None
    hashes: dict = None
    path_in_asset: str = None  # relative path of the data with respect to the query url

    def __init__(self, data: Any, url_object: StorageURL):
        """Initializes a standardized blob objects from data.
        Each storage plugin must provide their own blob by extending this class

        Parameters
        ----------
        data: any
            gcs Blob in case GCS, in case of aws or other storage, it could be a dictionary
        """
        # sanitize by removing any slashes from front and end
        self.hashes = {}
        self.initialize(data=data, url_object=url_object)
        # calculate path_in_asset, we need this to check for overrides
        # for example, user might enter 2 different urls and the query results could be such that
        # the path_in_asset of two objects collide. In that case, we consider only the first blob
        self.path_in_asset = os.path.relpath(self.name, url_object.dir_name)
        self.validate()

    def __eq__(self, other: StorageData):
        return self.path_in_asset == other.path_in_asset

    def __hash__(self):
        return hash(self.path_in_asset)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.path_in_asset)

    def validation_attributes(self) -> list:
        # note: even though hash is required, we are deferring hash validation
        # to the time of usage instead of time of instantiation
        # refer: self.get_hash for more details
        return ["name", "url", "host"]

    def validate(self):
        """subclass can override"""
        required_attrs = self.validation_attributes()
        for field in required_attrs:
            if not hasattr(self, field) or not getattr(self, field):
                raise exceptions.InvalidObjectSourceError(f"blob attribute {field} can not be null")

    @abc.abstractmethod
    def initialize(self, data: Any, url_object: StorageURL):
        """assign all attributes here, subclass must override

        Parameters
        ----------
        data: Any
            GcsBlob, S3Object, S3ObjectSummary, dict etc.
        url_object: StorageURL
            we need this to be able to stitch together the full url for the object

        Returns
        -------

        """
        raise NotImplementedError

    def get_hash(self) -> tuple:
        """returns the hash for the blob object

        Returns
        -------
        tuple:
            (hash_type, hash_value)
        """
        self.hashes = self.hashes or {}
        for hash_type in self.get_hash_preferences():
            if hash_type in self.hashes:
                return hash_type, self.hashes[hash_type]

        hash_type, hash_value = self.compute_hash()
        # verify
        if not hash_type or not hash_value:
            raise exceptions.InvalidObjectSourceError("hash_type and hash_value are both required")
        return self.validate_hash(hashes={hash_type: hash_value})

    def validate_hash(self, hashes: dict) -> tuple:
        """validates the blob's hash

        Parameters
        ----------
        hashes: dict
            {hash_type: hash_value}

        Returns
        -------
        tuple:
            (hash_type, hash_value) - validated
        """
        hash_prefs: list = self.get_hash_preferences()
        # order the hashes based on preferences
        ordered_hashes = []
        for hash_type in hash_prefs:
            if hash_type in hashes:
                ordered_hashes.append((hash_type, hashes[hash_type]))

        if not ordered_hashes:
            raise exceptions.InvalidObjectSourceError(f"invalid hash_type, allowed hashes are: {','.join(hash_prefs)}")
        # return the more preferred item
        return ordered_hashes[0]

    def compute_hash(self) -> tuple:
        """calculate and return hash
        subclass to implement

        Returns
        -------
        tuple:
            (hash_type, hash_value)

        """
        raise NotImplementedError

    def get_hash_preferences(self) -> list:
        # order of preferences for hash_type, subclass can override
        return ["md5", "crc32", "crc32c", "sha256", "sha1"]

    def compare_hash(self, other: StorageData) -> bool:
        """compare hashes of two blobs

        Parameters
        ----------
        other: StorageData
            other blob

        Returns
        -------
        bool:
            True if the blobs have same hashes, False otherwise
        """
        hash_type, hash_value = self.get_hash()
        # Look for the same hash_type in the other blob
        if hash_type in other.hashes:
            return hash_value == other.hashes[hash_type]
        return False

    def to_dict(self) -> dict:
        """Used in the server to serialize the object."""
        return {
            "url": self.url,
            "host": self.host,
            "name": self.name,
            "size": self.size,
            "content_type": self.content_type,
            "hashes": self.hashes,
            "path_in_asset": self.path_in_asset
        }



import os

from asset_utils.utils import utils
from .storage_url import StorageURL, PATH_SEPARATOR

HOST_SEPARATOR = "://"


class BlobStoreURL(StorageURL):
    host: str = None
    bucket: str = None
    path: str = None  # path to search for in bucket
    dir_name: str = None  # we need the dir_name to calculate the relative path for objects
    pattern: str = None  # for selecting a subset of files

    def parse_url(self, url: str) -> None:
        """parses a blob's url into:
        - storage i.e. gs, s3 or azure etc
        - bucket_name
        - blob_name

        Parameters
        ----------
        url: str
            can also include glob patterns

        Returns
        -------
        None

        """
        if not url:
            return

        if HOST_SEPARATOR not in url:
            self.log.info(f"not found: {url} doesn't contain {HOST_SEPARATOR}")
            return
        self.host, blob_url = url.split(HOST_SEPARATOR)
        if not blob_url:
            return
        idx = blob_url.find(PATH_SEPARATOR)
        if idx == -1:
            self.bucket = blob_url  # last part is the bucket
            return
        if idx > 0:
            self.bucket = blob_url[0:idx]
        if idx < len(blob_url) - 1:
            self.path = blob_url[idx + 1:]
        # check if there is pattern
        if self.path:
            pattern = utils.find_pattern(self.path)
            if pattern:
                # prefix is the parent directory
                prefix = self.path[:len(self.path) - len(pattern)]
                self.pattern = prefix + pattern  # need the full path string for fnmatch to work
                self.path = prefix

            dir_name = os.path.dirname(self.path)
            # sanitize by remove any leading or trailing slashes
            if dir_name:
                self.dir_name = dir_name.strip(PATH_SEPARATOR)

    def is_valid(self):
        return bool(self.host and self.bucket)  # path is optional for listing

    def is_remote(self):
        return self.is_valid()

    def url_for_blob(self, host: str, bucket: str, name: str):
        """construct the url for a blob object"""
        # sanitize the name by removing leading/trailing slashes
        return f"{host}{HOST_SEPARATOR}{bucket}{PATH_SEPARATOR}{name.strip(PATH_SEPARATOR)}"



import abc
import os
from typing import Type, Union

from asset_pluggy.plugin.object_content import ObjectContent
from asset_pluggy.storage.blob import StorageData
from asset_pluggy.storage.mount_config import MountConfig
from asset_pluggy.storage.storage_credentials import StorageCredentials
from asset_pluggy.storage.transporter import Transporter
from asset_pluggy.storage.urls import StorageURL
from asset_utils.common import Singleton, exceptions
from asset_utils.utils.log_utils import LoggingMixin


class AssetStorage(Singleton, LoggingMixin):
    prefixes: [str] = []  # list of prefixes
    name: str = None

    def post_init(self):
        self.validate()

    def validate(self):
        # verify transporter (test if it's working)
        transporter = self.get_transporter()
        if not transporter or not isinstance(transporter, Transporter):
            raise exceptions.InvalidStorageBackendError("StorageBackend must provide a Transporter")
        # cleanup transporter
        transporter.de_init()

    @property
    def credentials(self) -> dict:
        return StorageCredentials.shared().credentials

    @property
    def mount_config(self) -> MountConfig:
        return getattr(self, "_mount_config", None)

    @mount_config.setter
    def mount_config(self, value: str):
        self._mount_config = value

    @abc.abstractmethod
    def get_storage_url(self, url_string: str, ignore: str = None) -> StorageURL:
        raise NotImplementedError

    @abc.abstractmethod
    def allows_proxy(self):
        """allows objects to be created from its url as proxy"""
        raise NotImplementedError

    @abc.abstractmethod
    def allows_object_add(self):
        """allows objects to be created from its url"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_blob(self, url_string: str) -> StorageData:
        raise NotImplementedError

    @abc.abstractmethod
    def blob_exists(self, url_string: str) -> bool:
        """check if the blob exists without downloading data"""
        raise NotImplementedError

    @abc.abstractmethod
    def list_blobs(self, url: Union[StorageURL, str], ignore: str = None) -> [StorageData]:
        raise NotImplementedError

    @abc.abstractmethod
    def delete_blobs(self, url_strings: [str]) -> None:
        """delete one or more blobs"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_transporter(self) -> Transporter:
        raise NotImplementedError

    @abc.abstractmethod
    def get_content_class(self) -> Type[ObjectContent]:
        raise NotImplementedError

    def parse_blobs_from_url(self,
                             repo_dir: str,
                             url_string: str,
                             ignore: str = None,
                             dest_dir: str = None) -> [tuple]:
        """parse the source of content creation

        Parameters
        ----------
        repo_dir: str
            Repo directory
        url_string: StorageURL
            url_object from which the blobs were collected
        ignore: str
            glob pattern
        dest_dir: str
            optional, if provided, a new dest_dir is added to the file path

        Returns
        -------
        [tuple]:
            ([(blob: StorageData, path: str)])

        """
        search_url: StorageURL = self.get_storage_url(url_string=url_string, ignore=ignore)
        blobs: [StorageData] = self.list_blobs(url=search_url)
        sources = []
        for blob in blobs:
            content_dst = self.get_object_path(asset_root=repo_dir, blob=blob, parent_url=search_url)
            if dest_dir:
                content_dst = os.path.join(dest_dir, content_dst)
            sources.append((blob, content_dst))
        return sources

    @abc.abstractmethod
    def get_object_path(self, asset_root: str, blob: StorageData, parent_url: StorageURL) -> str:
        """Returns the path where the blob would be mounted inside the asset

        Parameters
        ----------
        asset_root: str
            Repo directory path
        blob: StorageData
        parent_url: StorageURL

        Returns
        -------
        str:
            mounting path inside the asset

        """
        raise NotImplementedError

    @abc.abstractmethod
    def url_is_file(self, url: Union[StorageURL, str]):
        raise NotImplementedError

    # used in asset-server
    @abc.abstractmethod
    def signed_url_for_blob(self, blob_url: str):
        raise NotImplementedError

    # used in asset-server
    @abc.abstractmethod
    def set_bucket_cors(self, bucket_url: str, origin_url):
        raise NotImplementedError

    # used in asset-server
    @abc.abstractmethod
    def get_bucket_cors(self, bucket_url: str):
        raise NotImplementedError



from abc import ABC

from .storage_data import StorageData


class BlobData(StorageData, ABC):
    url: str = None
    host: str = None
    bucket: str = None
    name: str = None
    hashes: dict = None  # hash_type is key, hash_value is value
    size: int = None
    content_type: str = None
    is_file: bool = None

    def to_dict(self) -> dict:
        """Used in the server to serialize the object."""
        data = super().to_dict()
        data.update({
            "url": self.url,
            "host": self.host,
            "bucket": self.bucket,
            "name": self.name,
            "hashes": self.hashes,
            "size": self.size,
            "content_type": self.content_type,
            "is_file": self.is_file
        })
        return data



import abc
from typing import Callable

from asset_utils.utils.log_utils import LoggingMixin

PATH_SEPARATOR = "/"


class StorageURL(LoggingMixin, metaclass=abc.ABCMeta):
    host: str = None
    url: str = None
    ignore: str = None
    # we need the dir_name to calculate the relative path for objects
    # every subclass must provide this value
    dir_name: str = None
    callback: Callable = None

    def __init__(self, url: str, ignore: str = None, callback: Callable = None):
        self.url = url
        self.ignore = ignore
        self.callback = callback
        self.parse_url(url=self.url)

    def execute_callback(self, *args):
        if self.callback:
            self.callback(*args)

    @abc.abstractmethod
    def parse_url(self, url: str) -> None:
        """parses a blob's url into:
        - storage i.e. gs, s3 or azure etc
        - bucket_name
        - blob_name

        Parameters
        ----------
        url: str

        Returns
        -------
        None

        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_valid(self):
        """verify if the url is sufficient to fetch data"""
        raise NotImplementedError

    @abc.abstractmethod
    def url_for_blob(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def is_remote(self):
        raise NotImplementedError



# No usage of ObjectManager found in the project.
from typing import Dict, Type

from asset_pluggy.plugin.asset_object import AssetObject
from asset_pluggy.plugin.plugin_manager import PluginManager
from asset_utils.common import exceptions, Singleton


class ObjectManager(Singleton):
    object_providers: Dict[str, Type[AssetObject]]

    def post_init(self):
        self.object_providers = {}
        self._init_providers()

    def _init_providers(self):
        plugin_manager = PluginManager()
        providers = plugin_manager.hook.asset_object_get()

        for provider in providers:
            prev_provider = self.object_providers.get(provider.unit, None)
            if prev_provider:
                raise exceptions.AssetObjectConflictException(provider, prev_provider)
            else:
                self.object_providers[provider.unit] = provider



from __future__ import annotations

import os.path
from typing import Callable

from asset_utils.common import exceptions
from asset_utils.utils.file_utils import FileUtils


class TransportResource:

    def __init__(self,
                 src: str,
                 dst: str,
                 hash: tuple = None,
                 callback: Callable = None,
                 **kwargs):
        self.src = src
        self.dst = dst
        self.src_hash = hash  # (hash_type, hash_value)
        self.callback = callback

    def __eq__(self, other: TransportResource):
        return self.src == other.src

    def __hash__(self):
        return hash(self.src)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.src)

    @property
    def dst_hash(self):
        try:
            return self._dst_hash
        except AttributeError:
            self._dst_hash = self.compute_dest_hash(hash_type=self.src_hash[0])
            return self._dst_hash

    @dst_hash.setter
    def dst_hash(self, x: tuple):
        self._dst_hash = x

    def validate(self) -> None:
        """The base implementation assumes files are available locally.
        Subclasses must override and provide their own implementation
        """
        if not os.path.exists(self.src):
            raise exceptions.InvalidObjectSourceError(f"file not found: {self.src}")

    def on_transfer_complete(self, *args):
        if self.callback:
            self.callback(*args)

    def verify_checksum(self):
        if not self.src_hash:
            return False
        ht, hv = self.dst_hash
        if ht != self.src_hash[0] or hv != self.src_hash[1]:
            return False
        return True

    def compute_dest_hash(self, hash_type: str):
        """subclass can override"""
        return FileUtils.file_hash(abs_path=self.dst, hash_type=hash_type)



from abc import ABC
from typing import Type

from asset_pluggy.plugin import hook_impl
from asset_pluggy.plugin.plugin_manager import PluginManager
from asset_pluggy.plugin.specs import AssetStorage
from asset_pluggy.plugin.storage_manager import StorageManager


def test_storage_manager():
    class AwsStorageTest(AssetStorage, ABC):
        prefixes = ["s3-test"]

    class GcsStorageTest(AssetStorage, ABC):
        prefixes = ["gs-test"]

    class GcrStorageTest(AssetStorage, ABC):
        prefixes = ["gcr.io-test", "us.gcr.io-test"]

    class AwsStorageTestPlugin:
        @hook_impl
        def asset_storage_get(self) -> Type[AssetStorage]:
            return AwsStorageTest

    class GcsStorageTestPlugin:
        @hook_impl
        def asset_storage_get(self) -> Type[AssetStorage]:
            return GcsStorageTest

    class GcrStorageTestPlugin:
        @hook_impl
        def asset_storage_get(self) -> Type[AssetStorage]:
            return GcrStorageTest

    plm = PluginManager.shared()
    plm.register(AwsStorageTestPlugin())
    plm.register(GcsStorageTestPlugin())
    plm.register(GcrStorageTestPlugin())
    sm = StorageManager.shared()
    providers: dict = sm.get_providers()
    assert len(providers) == 4
    expected = {
        "s3-test": AwsStorageTest,
        "gs-test": GcsStorageTest,
        "gcr.io-test": GcrStorageTest,
        "us.gcr.io-test": GcrStorageTest
    }
    for key in expected:
        actual = providers[key]
        exp = expected[key]
        assert actual == exp
        for prefix in exp.prefixes:
            assert prefix in actual.prefixes



import json
from unittest.mock import patch

import pytest

from asset_pluggy.plugin.storage_manager import StorageManager
from asset_pluggy.storage.asset_storage import AssetStorage
from asset_pluggy.storage.mount_config import MountConfig
from asset_pluggy.storage.storage_factory import StorageFactory
from asset_utils.common import exceptions


def test_storage_with_prefix():
    # this should succeed if installed
    prefixes = ["file://", "s3://", "gs://", "gcr.io/", "us.gcr.io/"]
    for prefix in prefixes:
        with pytest.raises(exceptions.InvalidStorageBackendError) as e:
            storage = StorageFactory.storage_with_prefix(prefix)
            assert issubclass(storage.__class__, AssetStorage)
            assert prefix in storage.prefixes
        assert str(e.value.args[0]) == f"{prefix} backend is not installed"

    # these should throw error
    prefixes = ["file", "s3", "gs", "gcr.io", "us.gcr.io"]
    for prefix in prefixes:
        with pytest.raises(exceptions.InvalidStorageBackendError) as e:
            _ = StorageFactory.storage_with_prefix(prefix)
        assert str(e.value.args[0]) == f"{prefix} backend is not installed"


def test_storage_for_url():
    # these should work once all plugins are installed
    urls = [
        ("gs://bucket/test/files/sample.yml", "gs://"),
        ("s3://bucket/test/test_data/sample.yml", "s3://"),
        ("gcr.io/project/image:my_tag", "gcr.io/"),
        ("folder/subfolder/sample.yml", "file://")
    ]
    for url, prefix in urls:
        with pytest.raises(exceptions.InvalidStorageBackendError) as e:
            storage = StorageFactory.storage_for_url(src_url=url)
            assert prefix in storage.prefixes
        assert str(e.value.args[0])


class MockStorage(AssetStorage):
    prefixes = ["s3://", "gs://"]


class MockStorageManager(StorageManager):
    _storage_providers = {
        "s3://": MockStorage,
        "gs://": MockStorage,
    }


@pytest.fixture(scope="module")
def mock_storage_manager():
    with patch('asset_pluggy.storage.storage_factory.StorageManager') as mock_manager:
        mock_manager.shared.return_value = MockStorageManager.shared()
        yield mock_manager


@pytest.fixture(scope="function")
def mock_environment(monkeypatch):
    mock_configs = {"s3://s3_bucket": "/mnt/s3",
                    "gs://gs_bucket": "/mnt/gs"}

    monkeypatch.setenv("ASSET_BUCKET_MT_CONFIG", json.dumps(mock_configs))


def test_mount_configs(mock_environment):
    configs = StorageFactory.mount_configs()
    assert len(configs) == 2
    assert isinstance(configs[0], MountConfig)
    assert configs[0].mount == "/mnt/s3"
    assert configs[1].url == "gs://gs_bucket"


def test_mount_configs_no_env(monkeypatch):
    monkeypatch.delenv("ASSET_BUCKET_MT_CONFIG", raising=False)
    assert StorageFactory.mount_configs() is None


@pytest.mark.parametrize("url, expected_mount", [
    ("s3://s3_bucket/file.txt", "/mnt/s3"),
    ("gs://gs_bucket/file.txt", "/mnt/gs"),
    ("/mnt/s3/file.txt", "/mnt/s3"),
    ("/mnt/gs/file.txt", "/mnt/gs"),
    ("http://example.com/file.txt", None),
])
def test_mount_config_for_url(mock_environment, url, expected_mount):
    config = StorageFactory.mount_config_for_url(url)
    if expected_mount:
        assert config.mount == expected_mount
    else:
        assert config is None


@pytest.mark.parametrize("url, expected_prefix", [
    ("s3://s3_bucket/file.txt", "s3://"),
    ("gs://gs_bucket/file.txt", "gs://"),
    ("/mnt/s3/file.txt", "s3://"),
    ("/mnt/gs/file.txt", "gs://")
])
def test_storage_for_mounted_url(mock_environment, url, expected_prefix):
    StorageManager.shared()._storage_providers = {
        "s3://": MockStorage,
        "gs://": MockStorage,
    }
    storage = StorageFactory.storage_for_url(url)
    if expected_prefix:
        assert isinstance(storage, MockStorage)
        assert expected_prefix in storage.prefixes
        assert storage.mount_config is not None
    else:
        # Assuming default_storage returns None or raises an exception
        assert storage is None



import pytest

from asset_pluggy.storage.mount_config import MountConfig
from asset_utils.common.exceptions import AssetException


@pytest.fixture
def mount_config():
    return MountConfig(mount="/mnt/data", url="s3://bucket/data")


def test_init(mount_config):
    assert mount_config.mount == "/mnt/data"
    assert mount_config.url == "s3://bucket/data"


def test_str_repr(mount_config):
    assert str(mount_config) == "MountConfig(mount=/mnt/data, url=s3://bucket/data)"
    assert repr(mount_config) == "MountConfig(mount=/mnt/data, url=s3://bucket/data)"


def test_eq_hash(mount_config):
    config1 = MountConfig("/mnt/data", "s3://bucket/data")
    config2 = MountConfig("/mnt/data", "s3://bucket/data")
    config3 = MountConfig("/mnt/other", "s3://bucket/other")

    assert config1 == config2
    assert config1 != config3
    assert hash(config1) == hash(config2)
    assert hash(config1) != hash(config3)


def test_to_dict(mount_config):
    expected = {"mount": "/mnt/data", "url": "s3://bucket/data"}
    assert mount_config.to_dict() == expected


def test_url_to_posix_valid(mount_config):
    url = "s3://bucket/data/file.txt"
    expected = "/mnt/data/file.txt"
    assert mount_config.url_to_posix(url) == expected


def test_url_to_posix_invalid(mount_config):
    invalid_url = "s3://other-bucket/data/file.txt"
    with pytest.raises(AssetException):
        mount_config.url_to_posix(invalid_url)


def test_posix_to_url_valid(mount_config):
    posix = "/mnt/data/file.txt"
    expected = "s3://bucket/data/file.txt"
    assert mount_config.posix_to_url(posix) == expected


def test_posix_to_url_invalid(mount_config):
    invalid_posix = "/other/data/file.txt"
    with pytest.raises(AssetException):
        mount_config.posix_to_url(invalid_posix)


def test_is_posix(mount_config):
    assert mount_config.is_posix("/mnt/data/file.txt") is True
    assert mount_config.is_posix("/other/data/file.txt") is False


def test_is_url(mount_config):
    assert mount_config.is_url("s3://bucket/data/file.txt") is True
    assert mount_config.is_url("s3://other-bucket/data/file.txt") is False


@pytest.mark.parametrize("mount,url", [
    ("/mnt/data", "s3://bucket/data"),
    ("/mnt/logs", "s3://logs-bucket/"),
    ("/data", "https://data.example.com"),
])
def test_multiple_configs(mount, url):
    config = MountConfig(mount, url)
    assert config.mount == mount
    assert config.url == url
    assert config.is_posix(f"{mount}/file.txt") is True
    assert config.is_url(f"{url}/file.txt") is True



from typing import Type

from asset_pluggy.plugin import hook_spec
from asset_pluggy.plugin.asset_object import AssetObject
from asset_pluggy.storage.asset_storage import AssetStorage


class StoragePlugin:
    """Storage plugin interface which can be used to discover registered storage backends    """

    @hook_spec
    def asset_storage_get(self) -> Type[AssetStorage]:
        """provides a storage backend to be used for storing asset data
        :return: AssetStorage compatible subclass
        """
        pass


class ObjectPlugin:
    @hook_spec
    def asset_object_get(self) -> Type[AssetObject]:
        """provides a Object type for any custom handling of data being added to asset"""
        pass


class PreCommitPlugin:
    """pre-commit rules for asset"""

    @hook_spec
    def asset_pre_commit_get(self) -> None:
        raise NotImplementedError


class ValidationPlugin:
    """asset validation rules"""

    @hook_spec
    def asset_validators_get(self) -> None:
        raise NotImplementedError


__all__ = [StoragePlugin, ObjectPlugin]



import json
import os
from typing import Type

from asset_pluggy.plugin.storage_manager import StorageManager
from asset_pluggy.storage.asset_storage import AssetStorage
from asset_pluggy.storage.mount_config import MountConfig
from asset_utils.common import exceptions


class StorageFactory:
    _instances = {}
    _storages = {}

    @classmethod
    def storage_with_prefix(cls, prefix: str) -> AssetStorage:
        storage = cls._storages.get(prefix, None)
        if not storage:
            storage = cls._add_storage(prefix=prefix)
        return storage

    @classmethod
    def _add_storage(cls, prefix: str):
        klass: Type[AssetStorage] = StorageManager.shared().get_storage(prefix=prefix)
        if not klass:
            raise exceptions.InvalidStorageBackendError(msg=f"{prefix} backend is not installed")
        storage = klass.shared()
        cls._storages[prefix] = storage
        return storage

    @classmethod
    def storage_for_url(cls, src_url: str):
        # translate to a standard blob url
        mt_cfg = cls.mount_config_for_url(src_url)
        if mt_cfg and mt_cfg.is_posix(src_url):
            # user might have passed posix
            norm_url = mt_cfg.posix_to_url(src_url)
        else:
            norm_url = src_url

        providers: dict = StorageManager.shared().get_providers()
        for prefix, storage_klass in providers.items():
            if norm_url.startswith(prefix):
                storage = cls.storage_with_prefix(prefix=prefix)
                if mt_cfg:
                    storage.mount_config = mt_cfg
                return storage
        return cls.default_storage()

    @classmethod
    def mount_configs(cls):
        cfgs = os.getenv("ASSET_BUCKET_MT_CONFIG")
        if cfgs:
            return [MountConfig(url=url, mount=mount) for url, mount in json.loads(cfgs).items()]
        return None

    @classmethod
    def mount_config_for_url(cls, src_url: str):
        mounts = cls.mount_configs()
        if mounts:
            for mount in mounts:
                # user can pass either posix or s3:// url
                if mount.is_posix(src_url) or mount.is_url(src_url):
                    return mount
        return None

    @classmethod
    def parse_sources(cls,
                      repo_dir: str,
                      targets: list,
                      proxy: bool = False,
                      dest_dir: str = None,
                      ignore: str = None) -> dict:
        """
        parses the targets which includes
         - collecting all files if the target is a directory
         - collecting all blobs/urls if the target is a gsurl or web url
        Parameters
        ----------
        repo_dir: str
            repo directory
        targets: [str]
            list of files, urls etc.
        proxy: bool
            if to be added as proxy
        dest_dir: str
            optional, to create a directory and add files in it
        ignore: str

        Returns
        -------
        dict:
            {storage_name: set<ContentSource>()}
        """
        result = {}
        for target in targets:
            storage: AssetStorage = cls.storage_for_url(src_url=target)
            if not storage.allows_object_add():
                raise exceptions.InvalidObjectSourceError(
                    f"you can not directly add objects from this source: {storage.name}")
            if proxy and not storage.allows_proxy():
                raise exceptions.InvalidObjectSourceError(f"proxy is not allowed for: {storage.name}")
            sources = storage.parse_blobs_from_url(repo_dir=repo_dir,
                                                   url_string=target,
                                                   ignore=ignore,
                                                   dest_dir=dest_dir)
            # fuse-mounted directories have an empty_file with trailing slash, we can't add these as assets
            sources = [item for item in sources if not str(item[0].name).endswith("/")]
            stored: dict = result.get(storage.name, {})
            for blob, path in sources:
                stored[path] = blob
            result[storage.name] = stored

        return result

    @classmethod
    def storage_with_name(cls, name: str):
        for storage_klass in cls._storages.values():
            if storage_klass.name == name:
                return storage_klass
        # check in storage factory
        klass: Type[AssetStorage] = StorageManager.shared().get_storage(prefix=None, name=name)
        if klass:
            return cls._add_storage(prefix=klass.prefixes[0])
        else:
            return cls.default_storage()

    @classmethod
    def default_storage(cls):
        return cls.storage_with_prefix(prefix=cls.default_storage_prefix())

    @classmethod
    def default_storage_prefix(cls) -> str:
        return "file://"  # Posix



from asset_utils.common.singleton import Singleton


class StorageCredentials(Singleton):

    def post_init(self):
        # default credentials for fetching asset and its contents i.e. files
        self._content_credentials: dict = {}
        # credentials for fetching only asset
        self._project_credentials: dict = {}
        self.use_content_credentials = False

    def set_content_credentials(self, cred: dict):
        """"always use for asset fetching only for most cases _content_credentials are same as _project_credentials"""
        self._content_credentials = cred

    def set_credentials(self, cred: dict):
        """always use for contents fetching only
         - credentials and asset_credentials could be same
         - only proxy assets, this would be different
        """
        self._project_credentials = cred

    @property
    def credentials(self) -> dict:
        """Return credentials based on the use_content_credentials flag"""
        if self.use_content_credentials:
            return self._content_credentials
        return self._project_credentials



from asset_pluggy.storage.storage_utils import filter_blobs


class MockBlob:
    def __init__(self, name):
        self.name = name


def test_filter_blobs():
    sample_blobs = [MockBlob(name="blob1.txt"),
                    MockBlob(name="ignore_blob.txt"),
                    MockBlob(name="blob2.txt")]

    # without ignore or pattern
    filtered = filter_blobs(blobs=sample_blobs, name_key="name")
    assert len(filtered) == 3, "Should return all blobs when no ignore or pattern is specified"

    # with ignore pattern
    filtered = filter_blobs(blobs=sample_blobs, name_key="name", ignore=["ignore_*"])
    assert len(filtered) == 2, "Should filter out blobs matching the ignore pattern"
    assert all(blob.name != "ignore_blob.txt" for blob in filtered), "Ignored blobs should not be in the result"

    # with match pattern
    filtered = filter_blobs(blobs=sample_blobs, name_key="name", pattern="*.txt")
    assert len(filtered) == 3, "Should return all blobs matching the txt pattern"

    # with ignore and match patterns
    filtered = filter_blobs(blobs=sample_blobs, name_key="name", ignore=["ignore_*"], pattern="*.txt")
    assert len(filtered) == 2, "Should return blobs matching the pattern but not the ignore pattern"
    assert all(blob.name != "ignore_blob.txt" for blob in filtered), "Ignored blobs should not be in the result"



from abc import ABC
from typing import Type

from asset_pluggy.plugin import hook_impl
from asset_pluggy.plugin.plugin_manager import PluginManager
from asset_pluggy.plugin.specs import AssetStorage


def test_plugin_manager():
    class TestStorage(AssetStorage, ABC):
        prefix = "test-storage"

    class TestStoragePlugin:
        @hook_impl
        def asset_storage_get(self) -> Type[AssetStorage]:
            return TestStorage

    plm = PluginManager.shared()
    plm.register(TestStoragePlugin())
    storages = plm.hook.asset_storage_get()
    # checking singleton also
    plm2 = PluginManager.shared()
    assert plm2 == plm
    storages2 = plm2.hook.asset_storage_get()
    assert len(storages) == 1
    assert len(storages2) == 1
    for storge in storages:
        assert storge is TestStorage
        assert storge.prefix == "test-storage"

    for storge in storages2:
        assert storge is TestStorage
        assert storge.prefix == "test-storage"



import pluggy

APP_NAME = "asset-manager"

hook_impl = pluggy.HookimplMarker(APP_NAME)
"""Storage providers can use this hook to implement custom storage class
"""

hook_spec = pluggy.HookspecMarker(APP_NAME)
"""hook_spec is used to standardize the calling within asset-core and registering and discovering plugins
"""



from asset_pluggy.storage.urls import BlobStoreURL


def test_blob_store_url():
    urls = [
        "gs://bucket/test/client/sample_yamls/annotate_params.yaml",
        "s3://bucket/rtd/models/",
        "s3://bucket/rtd/models",
        "gs://",
        "s3://",
        "gs://bucket/",
        "s3://bucket/",
        "gs://bucket",
        "s3://bucket",
        # test with patterns
        "gs://bucket/test/client/sample_yamls/*.yaml",
        "gs://bucket/test/client/*/*.yaml"
    ]
    expected = [
        ("gs", "bucket", "test/client/sample_yamls/annotate_params.yaml", "test/client/sample_yamls", None),
        ("s3", "bucket", "rtd/models/", "rtd/models", None),
        ("s3", "bucket", "rtd/models", "rtd", None),
        ("gs", None, None, None, None),
        ("s3", None, None, None, None),
        ("gs", "bucket", None, None, None),
        ("s3", "bucket", None, None, None),
        ("gs", "bucket", None, None, None),
        ("s3", "bucket", None, None, None),
        ("gs", "bucket", "test/client/sample_yamls/", "test/client/sample_yamls", "test/client/sample_yamls/*.yaml"),
        ("gs", "bucket", "test/client/", "test/client", "test/client/*/*.yaml"),
    ]

    for idx, url in enumerate(urls):
        blob_url = BlobStoreURL(url=url)
        exp = expected[idx]
        assert blob_url.url == url
        assert blob_url.host == exp[0]
        assert blob_url.bucket == exp[1]
        assert blob_url.path == exp[2]
        print(blob_url.__dict__)
        assert blob_url.dir_name == exp[3]



import json

import pytest

from asset_pluggy.storage.transporter import Transporter

GCS_1 = {
    "credentials": {
        "project_id": "general_random1_gcs",
        "email": "general@general.com",
        "private_key_id": "e6dd5048-7be8-4c74-a786-2d0596eb635b"
    },
    "prefixes": ["gcs"]
}

GCS_2 = {
    "credentials": {
        "project_id": "random2_general_gcs",
        "email": "random2@general.com",
        "private_key_id": "90d8ca60-9027-4ad2-9f8c-dfdfc22e874c"
    },
    "prefixes": ["gcs"]
}

S3_1 = {
    "credentials": {
        "project_id": "s3_random2_general",
        "email": "random2@general.com",
        "private_key_id": "b61ebf28-d5ab-490f-9bb0-f3a466eb91cb"
    },
    "prefixes": ["s3"]
}

S3_2 = {
    "credentials": {
        "project_id": "s3_random2_general",
        "email": "random2@general.com",
        "private_key_id": "04757873-e93f-4b8f-bd4f-0fb92b2fc972"
    },
    "prefixes": ["s3"]
}


def test_transporter():
    assert_transporter(cred=GCS_1)
    # test with null credentials
    Transporter.de_init()
    with pytest.raises(Exception) as e:
        assert_transporter(cred={})
    assert e
    Transporter.de_init()
    # test credentials switch
    creds = [GCS_1, GCS_2]
    for cred in creds:
        assert_transporter(cred=cred)


def assert_transporter(cred: dict):
    transport = Transporter.shared(credentials=cred.get("credentials"),
                                   prefixes=cred.get("prefixes"),
                                   validate=True)
    assert transport and isinstance(transport, Transporter)
    assert json.dumps(transport.credentials) == json.dumps(cred.get("credentials"))



from typing import Dict, Type

from asset_pluggy.plugin.plugin_manager import PluginManager
from asset_pluggy.storage.asset_storage import AssetStorage
from asset_utils.common.exceptions import AssetException
from asset_utils.common.singleton import Singleton


class AssetStorageConflictException(AssetException):
    def __init__(self, prefix: str, storage, prev_storage):
        msg = f'Input type `{prefix}` provided by {storage.__name__} already registered by {prev_storage.__name__}'
        super().__init__(self, msg)


class StorageManager(Singleton):
    _storage_providers: Dict[str, Type[AssetStorage]]

    def post_init(self, **kwargs):
        self._storage_providers = {}
        self._init_providers()

    def _init_providers(self):
        storage_providers = PluginManager.shared().hook.asset_storage_get()

        for storage in storage_providers:
            for prefix in storage.prefixes:
                prev_storage = self._storage_providers.get(prefix, None)
                if prev_storage:
                    raise AssetStorageConflictException(prefix=prefix, storage=storage, prev_storage=prev_storage)
                else:
                    self._storage_providers[prefix] = storage

    def get_storage(self, prefix: str, name: str = None):
        if prefix:
            return self._storage_providers.get(prefix, None)
        if name:
            for storage in self._storage_providers.values():
                if storage.name == name:
                    return storage
            return None

    def get_providers(self):
        return self._storage_providers
